(** @author: Benoît Montagu <benoit.montagu@inria.fr>
    Copyright Inria 2019
*)

open SimplLib.Main
open Cmdliner

let analysis_kind_t =
  let open Arg in
  let open AnalysisKind in
  value
  & vflag IndependentAttribute [
    IndependentAttribute,
    info ["independent-attribute"]
      ~doc:"Performs an analysis that infers a relation between \
            the current environment and the output value. The independent \
            attribute abstraction is used during the analysis.";
  ]

let run_internal_tests_t = let open Arg in
  value
  & flag
  & info ["run-internal-tests"]
    ~doc:"Runs internal tests."

let dont_learn_from_patterns_t = let open Arg in
  value
  & flag
  & info ["dont-learn-from-patterns"]
    ~doc:"Disables the learning of cases for values being pattern matched against."

let strict_errors_t = let open Arg in
  value
  & flag
  & info ["strict-errors"]
    ~doc:"Makes errors fatal."

let unroll_fixpoints_t = let open Arg in
  value
  & opt int 0
  & info ["unroll-fixpoints"]
    ~doc:"How many unrollings of fixpoints must be performed by default."

let delay_widening_t = let open Arg in
  value
  & opt int 0
  & info ["delay-widening"]
    ~doc:"How many fixpoints iterations must be performed by default before widening is used."

let options_t : Options.t Term.t =
  let open Term in
  let open Options in
  const (fun run_internal_tests
          dont_learn_from_patterns
          strict_errors unroll_fixpoints delay_widening
          analysis_kind ->
      { run_internal_tests;
        learn_from_patterns = not dont_learn_from_patterns;
        strict_errors;
        unroll_fixpoints = max 0 unroll_fixpoints;
        delay_widening = max 0 delay_widening;
        analysis_kind;
      })
  $ run_internal_tests_t
  $ dont_learn_from_patterns_t
  $ strict_errors_t
  $ unroll_fixpoints_t
  $ delay_widening_t
  $ analysis_kind_t

let file_t = let open Arg in
  required
  & pos 0 (some string) None
  & info [] ~docv:"FILE"
    ~doc:"File to analyse"

let simpl_t = let open Term in
  const simpl_print
  $ options_t
  $ file_t

let simpl_info = let open Term in
  info "simpl"
    ~doc:"Correlation analyser for a simple functional language."

let () =
  Term.(exit @@ eval (simpl_t, simpl_info))
